==> vite-plugin-proxy-0.6.0.tgz: package/src/middlewares/proxy-frontend.js <==
import parseurl from 'parseurl'
import * as proxy from '../proxy.js'
import * as http from '../http.js'
import finalhandler from 'finalhandler'
import createError, { HttpError } from 'http-errors'
import logger from '../logger.js'
import fs from 'fs/promises'
import path from 'path'

function getFSPath (path) {
  const index = path.indexOf('node_modules')
  if (index < 0) return path
  return path.substr(0, 32)
}

function proxyRequest (req, res, { hostname, protocol, port, secure }) {
  return proxy.web(req, res, {
    hostname,
    protocol,
    port,
    headers: { 'proxied-by-ox-proxy': true },
    secure,
    onRes (req, res, proxyRes) {
      if (proxyRes.statusCode >= 400) {
        throw createError(proxyRes.statusCode, proxyRes.message || '')
      }
      return http.pipe(proxyRes, res)
    }
  })
}

export default function ({ frontends }) {
  const cache = new Map()

  return async (req, res, next) => {
    if (req.headers['proxied-by-ox-proxy']) return next()

    const url = parseurl(req)

    // five stage system
    // 1. if we have a cached path, we can directly request the file from that system
    // 2. if it starts with /@fs, we can forward the request to the dev-server with according allowed paths
    // 3. if we have multiple dev-servers, we can check if the file exists there in the root
    // 4. check, if a path is configured for frontends and try to find the file there
    // 5. otherwise, request them in order

    // stage 1
    if (cache.has(url.path)) {
      logger.debug(`Found a handler for ${url.path} in cache`)
      try {
        const { target: { hostname, protocol, port }, secure } = cache.get(url.path)
        await proxyRequest(req, res, { hostname, protocol, port, secure })
        logger.debug(`Delivered ${url.path} by a cached handler`)
        return
      } catch (err) {
        logger.warn(`The handler for ${url.path} could not be used anymore.`)
        cache.delete(url.path)
      }
    }

    // stage 2
    if (url.path.startsWith('/@fs/')) {
      for (const frontend of frontends) {
        if (!frontend?.config?.fs?.allow) continue
        const fsPath = url.path.substring(4)
        const allowed = frontend.config.fs.allow.some(allowed => fsPath.startsWith(allowed))
        if (!allowed) continue

        try {
          const { hostname, protocol, port } = frontend.target
          logger.info(`Proxy ${url.path} to frontend ${protocol}//${hostname}:${port}`)
          await proxyRequest(req, res, { hostname, protocol, port, secure: frontend.secure })
          cache.set(url.path, frontend)
          logger.debug(`Found ${url.path} at ${protocol}//${hostname}:${port}`)
          return
        } catch (err) {
          logger.debug(`Did not find ${url.path} at ${protocol}//${hostname}:${port}, try next.`)
        }
      }
    }

    // stage 3
    for (const frontend of frontends) {
      if (!frontend?.config?.root) continue
      const file = path.join(frontend.config.root, url.pathname === '/' ? '/index.html' : url.pathname)
      try {
        await fs.stat(file)
        const { hostname, protocol, port } = frontend.target
        await proxyRequest(req, res, { hostname, protocol, port, secure: frontend.secure })
        cache.set(url.path, frontend)
        logger.debug(`Found ${url.path} at ${protocol}//${hostname}:${port}`)
        return
      } catch (err) {
        logger.trace(`Could not find ${url.path} from ${frontend.target.href} at ${frontend.config.root}`)
      }
    }

    // stage 4
    for (const frontend of frontends) {
      if (!frontend.path) continue
      const isRegExp = frontend.path instanceof RegExp
      const shouldHandle = isRegExp ? frontend.path.test(url.path) : url.path.startsWith(frontend.path)
      if (!shouldHandle) continue
      const { hostname, protocol, port } = frontend.target
      logger.info(`Proxy ${url.path} to frontend ${protocol}//${hostname}:${port}`)
      try {
        await proxyRequest(req, res, { hostname, protocol, port, secure: frontend.secure })
        cache.set(url.path, frontend)
        logger.debug(`Found ${url.path} at ${protocol}//${hostname}:${port}`)
        return
      } catch (err) {
        logger.debug(`Did not find ${url.path} at ${protocol}//${hostname}:${port}, try next.`)
      }
    }

    // stage 5
    logger.debug(`Fall back to autofind ${url.path} in frontends`)
    const errors = []
    for (const frontend of frontends) {
      const { hostname, protocol, port } = frontend.target
      try {
        await proxyRequest(req, res, { hostname, protocol, port, secure: frontend.secure })
        cache.set(url.path, frontend)
        logger.debug(`Found ${url.path} at ${protocol}//${hostname}:${port}`)
        return
      } catch (err) {
        logger.debug(`Did not find ${url.path} at ${protocol}//${hostname}:${port}, try next.`)
        errors.push(err)
      }
    }

    const err = errors.reduce((memo, err) => {
      // keep the first non-http error and forward it
      if (!(memo instanceof HttpError)) return memo
      // otherwise, use the last http error
      return err
    }, new createError.NotFound())

    const handle = finalhandler(req, res)
    if (err instanceof HttpError) {
      // in case of http-error the proxy explicitedly received a error code
      logger.warn(`Could not find ${url.path} in frontends`)
      return handle(err)
    }
    logger.error(`Something went wrong, while resolving a file in frontends: ${err.stack}`)
    return handle()
  }
}

==> vite-plugin-proxy-0.6.1.tgz: package/src/middlewares/proxy-frontend.js <==
import parseurl from 'parseurl'
import * as proxy from '../proxy.js'
import * as http from '../http.js'
import finalhandler from 'finalhandler'
import createError, { HttpError } from 'http-errors'
import logger from '../logger.js'
import fs from 'fs/promises'
import path from 'path'

function getFSPath (path) {
  const index = path.indexOf('node_modules')
  if (index < 0) return path
  return path.substr(0, 32)
}

function proxyRequest (req, res, { hostname, protocol, port, secure }) {
  return proxy.web(req, res, {
    hostname,
    protocol,
    port,
    headers: { 'proxied-by-ox-proxy': true },
    secure,
    onRes (req, res, proxyRes) {
      if (proxyRes.statusCode >= 400) {
        throw createError(proxyRes.statusCode, proxyRes.message || '')
      }
      return http.pipe(proxyRes, res)
    }
  })
}

export default function ({ frontends }) {
  const cache = new Map()
  // this is used to remove the /@fs/ prefix from the path
  const fsPathLength = process.platform === 'win32' ? '/@fs/'.length : '/@fs'.length

  return async (req, res, next) => {
    if (req.headers['proxied-by-ox-proxy']) return next()

    const url = parseurl(req)

    // five stage system
    // 1. if we have a cached path, we can directly request the file from that system
    // 2. if it starts with /@fs, we can forward the request to the dev-server with according allowed paths
    // 3. if we have multiple dev-servers, we can check if the file exists there in the root
    // 4. check, if a path is configured for frontends and try to find the file there
    // 5. otherwise, request them in order

    // stage 1
    if (cache.has(url.path)) {
      logger.debug(`Found a handler for ${url.path} in cache`)
      try {
        const { target: { hostname, protocol, port }, secure } = cache.get(url.path)
        await proxyRequest(req, res, { hostname, protocol, port, secure })
        logger.debug(`Delivered ${url.path} by a cached handler`)
        return
      } catch (err) {
        logger.warn(`The handler for ${url.path} could not be used anymore.`)
        cache.delete(url.path)
      }
    }

    // stage 2
    if (url.path.startsWith('/@fs/')) {
      for (const frontend of frontends) {
        if (!frontend?.config?.fs?.allow) continue
        const fsPath = url.path.substring(fsPathLength)
        const allowed = frontend.config.fs.allow.some(allowed => fsPath.startsWith(allowed))
        if (!allowed) continue

        try {
          const { hostname, protocol, port } = frontend.target
          logger.info(`Proxy ${url.path} to frontend ${protocol}//${hostname}:${port}`)
          await proxyRequest(req, res, { hostname, protocol, port, secure: frontend.secure })
          cache.set(url.path, frontend)
          logger.debug(`Found ${url.path} at ${protocol}//${hostname}:${port}`)
          return
        } catch (err) {
          logger.debug(`Did not find ${url.path} at ${protocol}//${hostname}:${port}, try next.`)
        }
      }
    }

    // stage 3
    for (const frontend of frontends) {
      if (!frontend?.config?.root) continue
      const file = path.join(frontend.config.root, url.pathname === '/' ? '/index.html' : url.pathname)
      try {
        await fs.stat(file)
        const { hostname, protocol, port } = frontend.target
        await proxyRequest(req, res, { hostname, protocol, port, secure: frontend.secure })
        cache.set(url.path, frontend)
        logger.debug(`Found ${url.path} at ${protocol}//${hostname}:${port}`)
        return
      } catch (err) {
        logger.trace(`Could not find ${url.path} from ${frontend.target.href} at ${frontend.config.root}`)
      }
    }

    // stage 4
    for (const frontend of frontends) {
      if (!frontend.path) continue
      const isRegExp = frontend.path instanceof RegExp
      const shouldHandle = isRegExp ? frontend.path.test(url.path) : url.path.startsWith(frontend.path)
      if (!shouldHandle) continue
      const { hostname, protocol, port } = frontend.target
      logger.info(`Proxy ${url.path} to frontend ${protocol}//${hostname}:${port}`)
      try {
        await proxyRequest(req, res, { hostname, protocol, port, secure: frontend.secure })
        cache.set(url.path, frontend)
        logger.debug(`Found ${url.path} at ${protocol}//${hostname}:${port}`)
        return
      } catch (err) {
        logger.debug(`Did not find ${url.path} at ${protocol}//${hostname}:${port}, try next.`)
      }
    }

    // stage 5
    logger.debug(`Fall back to autofind ${url.path} in frontends`)
    const errors = []
    for (const frontend of frontends) {
      const { hostname, protocol, port } = frontend.target
      try {
        await proxyRequest(req, res, { hostname, protocol, port, secure: frontend.secure })
        cache.set(url.path, frontend)
        logger.debug(`Found ${url.path} at ${protocol}//${hostname}:${port}`)
        return
      } catch (err) {
        logger.debug(`Did not find ${url.path} at ${protocol}//${hostname}:${port}, try next.`)
        errors.push(err)
      }
    }

    const err = errors.reduce((memo, err) => {
      // keep the first non-http error and forward it
      if (!(memo instanceof HttpError)) return memo
      // otherwise, use the last http error
      return err
    }, new createError.NotFound())

    const handle = finalhandler(req, res)
    if (err instanceof HttpError) {
      // in case of http-error the proxy explicitedly received a error code
      logger.warn(`Could not find ${url.path} in frontends`)
      return handle(err)
    }
    logger.error(`Something went wrong, while resolving a file in frontends: ${err.stack}`)
    return handle()
  }
}
